@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -55,7 +55,7 @@
 #define STUNNEL_PLATFORM "Win32"
 #endif
 #define SERVICE_NAME TEXT("stunnel")
-#define SERVICE_DISPLAY_NAME TEXT("Stunnel SSL wrapper")
+#define SERVICE_DISPLAY_NAME TEXT("Stunnel TLS wrapper")
 #endif
 
 /* mingw-Patches-1825044 is missing in Debian Squeeze */
@@ -79,7 +79,7 @@ NOEXPORT int save_text_file(LPTSTR, char *);
 NOEXPORT void update_logs(void);
 NOEXPORT LPTSTR log_txt(void);
 
-NOEXPORT void daemon_thread(void *);
+NOEXPORT unsigned __stdcall daemon_thread(void *);
 
 NOEXPORT void valid_config(void);
 NOEXPORT void invalid_config(void);
@@ -137,14 +137,14 @@ static HANDLE main_initialized=NULL; /* global initialization performed */
 static HANDLE config_ready=NULL; /* reload without a valid configuration */
 static LONG new_logs=0;
 
-static UI_DATA *ui_data=NULL;
-
 static struct {
     char *config_file;
     unsigned service:1, install:1, uninstall:1, start:1, stop:1,
         quiet:1, exit:1, reload:1, reopen:1;
 } cmdline;
 
+static char ui_pass[PEM_BUFSIZE];
+
 /**************************************** initialization */
 
 int WINAPI WinMain(HINSTANCE this_instance, HINSTANCE prev_instance,
@@ -331,6 +331,7 @@ NOEXPORT int initialize_winsock() {
 /**************************************** GUI thread */
 
 NOEXPORT int gui_loop() {
+    HANDLE handle;
 #ifdef _WIN32_WCE
     WNDCLASS wc;
 #else
@@ -381,8 +382,12 @@ NOEXPORT int gui_loop() {
     /* auto-reset, non-signaled events */
     main_initialized=CreateEvent(NULL, FALSE, FALSE, NULL);
     config_ready=CreateEvent(NULL, FALSE, FALSE, NULL);
-    /* hwnd needs to be initialized before _beginthread() */
-    _beginthread(daemon_thread, DEFAULT_STACK_SIZE, NULL);
+    /* hwnd needs to be initialized before _beginthreadex() */
+    handle=(HANDLE)_beginthreadex(NULL, DEFAULT_STACK_SIZE,
+        daemon_thread, NULL, 0, NULL);
+    if(!handle)
+        fatal("Failed to create the daemon thread");
+    CloseHandle(handle);
     WaitForSingleObject(main_initialized, INFINITE);
     /* logging subsystem is now available */
 
@@ -693,11 +698,13 @@ NOEXPORT LRESULT CALLBACK pass_proc(HWND dialog_handle, UINT message,
         /* set the default push button to "Cancel" */
         SendMessage(dialog_handle, DM_SETDEFID, (WPARAM)IDCANCEL, (LPARAM)0);
 
-        key_file_name=str2tstr(ui_data->section->key);
-        titlebar=str_tprintf(TEXT("Private key: %s"), key_file_name);
-        str_free(key_file_name);
-        SetWindowText(dialog_handle, titlebar);
-        str_free(titlebar);
+        if(current_section) { /* should always be set */
+            key_file_name=str2tstr(current_section->key);
+            titlebar=str_tprintf(TEXT("Private key: %s"), key_file_name);
+            str_free(key_file_name);
+            SetWindowText(dialog_handle, titlebar);
+            str_free(titlebar);
+        }
         return TRUE;
 
     case WM_COMMAND:
@@ -722,9 +729,9 @@ NOEXPORT LRESULT CALLBACK pass_proc(HWND dialog_handle, UINT message,
                 (WPARAM)0 /* line 0 */, (LPARAM)pass_dialog.txt);
             pass_dialog.txt[pass_len]='\0'; /* null-terminate the string */
 
-            /* convert input passphrase to UTF-8 string (as ui_data->pass) */
+            /* convert input passphrase to UTF-8 string (as ui_pass) */
             pass_txt=tstr2str(pass_dialog.txt);
-            strcpy(ui_data->pass, pass_txt);
+            strcpy(ui_pass, pass_txt);
             str_free(pass_txt);
 
             EndDialog(dialog_handle, TRUE);
@@ -741,17 +748,21 @@ NOEXPORT LRESULT CALLBACK pass_proc(HWND dialog_handle, UINT message,
     UNREFERENCED_PARAMETER(lParam);
 }
 
-int passwd_cb(char *buf, int size, int rwflag, void *userdata) {
-    (void)rwflag; /* squash the unused parameter warning */
+int ui_passwd_cb(char *buf, int size, int rwflag, void *userdata) {
+    int len;
 
-    ui_data=userdata;
-    if(size<0) /* just in case */
-        return 0;
+    (void)rwflag; /* squash the unused parameter warning */
+    (void)userdata; /* squash the unused parameter warning */
     if(!DialogBox(ghInst, TEXT("PassBox"), hwnd, (DLGPROC)pass_proc))
-        return 0; /* error */
-    strncpy(buf, ui_data->pass, (size_t)size);
-    buf[size-1]='\0';
-    return (int)strlen(buf);
+        return 0; /* dialog cancelled or failed */
+    len=(int)strlen(ui_pass);
+    if(len<0 || size<0) /* the API uses signed integers */
+        return 0;
+    if(len>size) /* truncate the returned data if needed */
+        len=size;
+    memcpy(buf, ui_pass, (size_t)len);
+    memset(ui_pass, 0, sizeof ui_pass);
+    return len;
 }
 
 #ifndef OPENSSL_NO_ENGINE
@@ -770,14 +781,10 @@ UI_METHOD *UI_stunnel() {
 }
 
 NOEXPORT int pin_cb(UI *ui, UI_STRING *uis) {
-    ui_data=UI_get0_user_data(ui); /* was: ui_data=UI_get_app_data(ui); */
-    if(!ui_data) {
-        s_log(LOG_ERR, "INTERNAL ERROR: user data data pointer");
-        return 0;
-    }
     if(!DialogBox(ghInst, TEXT("PassBox"), hwnd, (DLGPROC)pass_proc))
-        return 0; /* error */
-    UI_set_result(ui, uis, ui_data->pass);
+        return 0; /* dialog cancelled or failed */
+    UI_set_result(ui, uis, ui_pass);
+    memset(ui_pass, 0, sizeof ui_pass);
     return 1;
 }
 #endif
@@ -891,7 +898,7 @@ NOEXPORT LPTSTR log_txt(void) {
 
 /**************************************** worker thread */
 
-NOEXPORT void daemon_thread(void *arg) {
+NOEXPORT unsigned __stdcall daemon_thread(void *arg) {
     (void)arg; /* squash the unused parameter warning */
 
     tls_alloc(NULL, NULL, "main"); /* new thread-local storage */
@@ -899,19 +906,20 @@ NOEXPORT void daemon_thread(void *arg) {
     SetEvent(main_initialized); /* unlock the GUI thread */
     /* get a valid configuration */
     while(main_configure(cmdline.config_file, NULL)) {
-        cmdline.config_file=NULL; /* don't retry commandline switches */
+        if(cmdline.config_file && *cmdline.config_file=='-')
+            cmdline.config_file=NULL; /* don't retry commandline switches */
         unbind_ports(); /* in case initialization failed after bind_ports() */
         log_flush(LOG_MODE_ERROR); /* otherwise logs are buffered */
         PostMessage(hwnd, WM_INVALID_CONFIG, 0, 0); /* display error */
         WaitForSingleObject(config_ready, INFINITE);
-        log_close(); /* prevent main_configure() from logging in error mode */
     }
     PostMessage(hwnd, WM_VALID_CONFIG, 0, 0);
 
     /* start the main loop */
     daemon_loop();
     main_cleanup();
-    _endthread(); /* SIGNAL_TERMINATE received */
+    _endthreadex(0); /* SIGNAL_TERMINATE received */
+    return 0;
 }
 
 /**************************************** helper functions */
@@ -992,7 +1000,7 @@ NOEXPORT void update_peer_menu(void) {
             TEXT("Peer certificate chain has been saved.\n")
             TEXT("Add the following lines to section [%s]:\n")
             TEXT("\tCAfile = peer-%s.pem\n")
-            TEXT("\tverify = 3\n")
+            TEXT("\tverifyPeer = yes\n")
             TEXT("to enable cryptographic authentication.\n")
             TEXT("Then reload stunnel configuration file."),
             servname, servname);
